@@ -18,12 +18,9 @@ from io import BytesIO
 from typing import List, Dict, Set, Optional, Any, Callable, Tuple, Iterable
 import docker
 from ... import Error
-from ..._util import (
-    chmod_R_plus,
-    PygtailLogger,
-)
+from ..._util import chmod_R_plus, TerminationSignalFlag
 from ..._util import StructuredLogMessage as _
-from .. import config, _statusbar
+from .. import config
 from ..error import Interrupted, Terminated
 from ..task_container import TaskContainer
 
@@ -37,14 +34,20 @@ class SwarmContainer(TaskContainer):
 
     @classmethod
     def global_init(cls, cfg: config.Loader, logger: logging.Logger) -> None:
-        client = docker.from_env(version="auto")
         worker_nodes = []
-        try:
+
+        with contextlib.ExitStack() as cleanup:
+            client = docker.from_env(version="auto")
+            cleanup.callback(lambda: client.close())
+            terminating = cleanup.enter_context(TerminationSignalFlag(logger))
             logger.debug("dockerd :: " + json.dumps(client.version())[1:-1])
 
             # initialize swarm
             state = "(unknown)"
             while True:
+                if terminating():
+                    raise Terminated()
+
                 info = client.info()
                 if "Swarm" in info and "LocalNodeState" in info["Swarm"]:
                     logger.debug(_("swarm info", **info["Swarm"]))
@@ -98,8 +101,6 @@ class SwarmContainer(TaskContainer):
                     "This is normal if other miniwdl processes are running concurrently; "
                     "otherwise, stale state could interfere with this run. To reset it, `docker swarm leave --force`"
                 )
-        finally:
-            client.close()
 
         # Detect swarm's CPU & memory resources. Even on a localhost swarm, these may be less than
         # multiprocessing.cpu_count() and psutil.virtual_memory().total; in particular on macOS,
@@ -232,13 +233,7 @@ class SwarmContainer(TaskContainer):
 
             # stream stderr into log
             with contextlib.ExitStack() as cleanup:
-                poll_stderr = cleanup.enter_context(
-                    PygtailLogger(
-                        logger,
-                        self.host_stderr_txt(),
-                        callback=self.stderr_callback,
-                    )
-                )
+                poll_stderr = cleanup.enter_context(self.poll_stderr_context(logger))
 
                 # poll for container exit
                 running_states = {"preparing", "running"}
@@ -277,12 +272,7 @@ class SwarmContainer(TaskContainer):
                         # indicate actual container start in status bar
                         # 'preparing' is when docker is pulling and extracting the image, which can
                         # be a lengthy and somewhat intensive operation, so we count it as running.
-                        cleanup.enter_context(
-                            _statusbar.task_running(
-                                self.runtime_values.get("cpu", 0),
-                                self.runtime_values.get("memory_reservation", 0),
-                            )
-                        )
+                        cleanup.enter_context(self.task_running_context())
                         was_running = True
                     if "running" in self._observed_states:
                         poll_stderr()
@@ -351,20 +341,6 @@ class SwarmContainer(TaskContainer):
         return image_tag
 
     def prepare_mounts(self, logger: logging.Logger) -> List[docker.types.Mount]:
-        def touch_mount_point(host_path: str) -> None:
-            # touching each mount point ensures they'll be owned by invoking user:group
-            assert host_path.startswith(self.host_dir + "/")
-            if host_path.endswith("/"):
-                os.makedirs(host_path, exist_ok=True)
-            else:
-                os.makedirs(os.path.dirname(host_path), exist_ok=True)
-                with open(host_path, "x") as _:
-                    pass
-            # providing g+rw on files (and g+rwx on directories) ensures the command will have
-            # permission to them regardless of which uid it runs as in the container (since we add
-            # the container to the invoking user's primary group)
-            chmod_R_plus(host_path.rstrip("/"), file_bits=0o660, dir_bits=0o770)
-
         def escape(s):
             # docker processes {{ interpolations }}
             return s.replace("{{", '{{"{{"}}')
@@ -393,7 +369,7 @@ class SwarmContainer(TaskContainer):
                     self.host_dir, os.path.relpath(container_path.rstrip("/"), self.container_dir)
                 )
                 if not os.path.exists(host_mount_point):
-                    touch_mount_point(
+                    self.touch_mount_point(
                         host_mount_point + ("/" if container_path.endswith("/") else "")
                     )
                 mounts.append(
@@ -413,7 +389,7 @@ class SwarmContainer(TaskContainer):
             )
         )
         # mount stdout, stderr, and working directory read/write
-        touch_mount_point(self.host_stdout_txt())
+        self.touch_mount_point(self.host_stdout_txt())
         mounts.append(
             docker.types.Mount(
                 escape(os.path.join(self.container_dir, "stdout.txt")),
@@ -421,7 +397,7 @@ class SwarmContainer(TaskContainer):
                 type="bind",
             )
         )
-        touch_mount_point(self.host_stderr_txt())
+        self.touch_mount_point(self.host_stderr_txt())
         mounts.append(
             docker.types.Mount(
                 escape(os.path.join(self.container_dir, "stderr.txt")),
@@ -436,8 +412,18 @@ class SwarmContainer(TaskContainer):
                 type="bind",
             )
         )
-        for p in [self.host_work_dir(), os.path.join(self.host_dir, "command")]:
-            chmod_R_plus(p, file_bits=0o660, dir_bits=0o770)
+
+        # providing g+rw on files (and g+rwx on directories) in the task directory ensures unix
+        # permissions won't block the command regardless of which uid it runs as in the container
+        # (since we add the container to the invoking user's primary group)
+        for mnt in [
+            os.path.join(self.host_dir, "command"),
+            self.host_stdout_txt(),
+            self.host_stderr_txt(),
+            self.host_work_dir(),
+        ]:
+            chmod_R_plus(mnt, file_bits=0o660, dir_bits=0o770)
+
         return mounts
 
     def misc_config(
@@ -475,6 +461,8 @@ class SwarmContainer(TaskContainer):
             logger.warning(
                 "container command will run as a root/wheel group member, since this is your primary group (gid=0)"
             )
+        if self.runtime_values.get("gpu", False):
+            logger.warning("ignored runtime.gpu (not yet implemented)")
         return resources, user, groups
 
     def poll_service(
