@@ -1,9 +1,11 @@
+# NOTE: this file is excluded from coverage analysis since alternate container backends may not be
+#       available in the CI environment. To test locally: prove -v tests/singularity.t
 import os
 import shlex
 import logging
 import tempfile
 import subprocess
-from typing import List, Tuple
+from typing import List, Tuple, Optional
 from contextlib import ExitStack
 from ...Error import InputError, RuntimeError
 from ..._util import StructuredLogMessage as _
@@ -16,6 +18,8 @@ class SingularityContainer(SubprocessBase):
     Singularity task runtime based on cli_subprocess.SubprocessBase
     """
 
+    image_cache_dir: Optional[str]
+
     @classmethod
     def global_init(cls, cfg: config.Loader, logger: logging.Logger) -> None:
         cmd = cfg.get_list("singularity", "exe") + ["--version"]
@@ -31,6 +35,14 @@ class SingularityContainer(SubprocessBase):
             raise RuntimeError(
                 f"Unable to check `{' '.join(cmd)}`; verify Singularity installation"
             )
+
+        image_cache_dir = cfg.get("singularity", "image_cache")
+        if image_cache_dir != "":
+            cls.image_cache_dir = os.path.abspath(image_cache_dir)
+            os.makedirs(cls.image_cache_dir, exist_ok=True)
+        else:
+            cls.image_cache_dir = None
+
         logger.notice(  # pyre-ignore
             _(
                 "Singularity runtime initialized (BETA)",
@@ -49,11 +61,16 @@ class SingularityContainer(SubprocessBase):
     def _pull_invocation(self, logger: logging.Logger, cleanup: ExitStack) -> Tuple[str, List[str]]:
         image, invocation = super()._pull_invocation(logger, cleanup)
         docker_uri = "docker://" + image
-
-        # The docker image layers are cached in SINGULARITY_CACHEDIR, so we don't need to keep the
-        # *.sif
-        pulldir = cleanup.enter_context(tempfile.TemporaryDirectory(prefix="miniwdl_sif_"))
-        return (docker_uri, self.cli_exe + ["pull", "--dir", pulldir, docker_uri])
+        pulldir = self.image_cache_dir or cleanup.enter_context(
+            tempfile.TemporaryDirectory(prefix="miniwdl_sif_")
+        )
+        image_name = docker_uri.replace("/", "_").replace(":", "_")
+        image_path = os.path.join(pulldir, image_name + ".sif")
+        if not os.path.exists(image_path):
+            return image_path, self.cli_exe + ["pull", image_path, docker_uri]
+        # If path already exists, no need to use a pull invocation.
+        logger.info(_("Singularity SIF found in image cache directory", sif=image_path))
+        return image_path, []
 
     def _run_invocation(self, logger: logging.Logger, cleanup: ExitStack, image: str) -> List[str]:
         """
@@ -77,7 +94,11 @@ class SingularityContainer(SubprocessBase):
         # Also create a scratch directory and mount to /tmp and /var/tmp
         # For context why this is needed:
         #   https://github.com/hpcng/singularity/issues/5718
-        tempdir = cleanup.enter_context(tempfile.TemporaryDirectory(prefix="miniwdl_singularity_"))
+        # TODO: provide opt-out for those able to edit /etc/singularity/singularity.conf to
+        #       increase sessiondir max size
+        tempdir = cleanup.enter_context(
+            tempfile.TemporaryDirectory(prefix="_singularity_tmpdir_", dir=self.host_dir)
+        )
         os.mkdir(os.path.join(tempdir, "tmp"))
         os.mkdir(os.path.join(tempdir, "var_tmp"))
         mounts.append(("/tmp", os.path.join(tempdir, "tmp"), True))
