@@ -280,6 +280,26 @@ class TestTasks(unittest.TestCase):
                 }
                 """)[0]
 
+    def test_issue571(self):
+        # regression test https://github.com/chanzuckerberg/miniwdl/issues/571
+        for version in ("1.0", "1.1", "development"):
+            task = WDL.parse_tasks("""
+                task echo {
+                    input {String name = "world"}
+                    command <<<
+                        echo hello, ~{name} >> out
+                        echo ">~{name}" >> out
+                        echo ">>~~{name}" >> out~
+                    ~~>>>
+                    output {String out = read_string("out")}
+                }
+            """, version=version)[0]
+            task.typecheck()
+            self.assertEqual(len(task.command.parts), 7, version)
+            self.assertIsInstance(task.command.parts[1], WDL.Expr.Placeholder, version)
+            self.assertIsInstance(task.command.parts[3], WDL.Expr.Placeholder, version)
+            self.assertIsInstance(task.command.parts[5], WDL.Expr.Placeholder, version)
+
     def test_meta(self):
         task = WDL.parse_tasks("""
         task wc {
@@ -2299,7 +2319,9 @@ class TestStruct(unittest.TestCase):
         self.assertEqual(len(ctx.exception.exceptions), 4)
         for i in range(4):
             self.assertTrue(isinstance(ctx.exception.exceptions[i], WDL.Error.StaticTypeMismatch))
-        self.assertEqual(str(ctx.exception.exceptions[2]), "Expected Person instead of object(age : Boolean, name : String)")
+        self.assertEqual(str(ctx.exception.exceptions[2]),
+            "Expected Person instead of object(age : Boolean, name : String);"
+            " type mismatch using Boolean to initialize Int age member of struct Person")
 
         doc = r"""
         version 1.0
@@ -2307,7 +2329,7 @@ class TestStruct(unittest.TestCase):
         workflow wf {
             Array[Person] ppl = [
                 object { name: 'alyssa', friends: [2,4] },
-                object { "name": "ben", 'friends': [8,16]},
+                object { "name": "ben", 'friends': [8,16], id: 42},
                 object { 'name': "cy", "friends": [32,64] }
             ]
         }
@@ -2315,6 +2337,7 @@ class TestStruct(unittest.TestCase):
         struct Person {
             String name
             Array[Int] friends
+            Int? id
         }
         """
         doc = WDL.parse_document(doc)
